<?php
/**
 * Copyright © Magento, Inc. All rights reserved.
 * See COPYING.txt for license details.
 */
declare(strict_types=1);

namespace FiloBlu\ExtMysqlSearch\Search\Adapter\Mysql\Query\Builder;

use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\DB\Helper\Mysql\Fulltext;
use Magento\Framework\DB\Select;
use Magento\Framework\Search\Adapter\Mysql\Field\FieldInterface;
use Magento\Framework\Search\Adapter\Mysql\Field\ResolverInterface;
use Magento\Framework\Search\Adapter\Mysql\ScoreBuilder;
use Magento\Framework\Search\Adapter\Preprocessor\PreprocessorInterface;
use Magento\Framework\Search\Request\Query\BoolExpression;
use Magento\Framework\Search\Request\QueryInterface as RequestQueryInterface;


/**
 * Class Match
 * @package FiloBlu\ExtMysqlSearch\Search\Adapter\Mysql\Query\Builder
 */
class Match extends \Magento\Framework\Search\Adapter\Mysql\Query\Builder\Match
{

    const SPECIAL_CHARACTERS_PATH = 'catalog/search/special_characters_field';

    const MINIMAL_CHARACTER_LENGTH_PATH = 'catalog/search/special_characters_length_field';

    /**
     *
     * @var string
     */
    protected $specialCharacters;
    /**
     * @var int
     */
    protected $minimalCharacterLength;

    /**
     * @var ScopeConfigInterface
     */
    private $config;

    protected $symbols= [];
    /**
     * @var ResolverInterface
     */
    private $resolverInterface;
    /**
     * @var Fulltext
     */
    private $fullTextHelper;

    /**
     * @var string
     */
    private $fullTextSearchMode;

    /**
     * @var array
     */
    protected $preprocessors;


    /**
     * Match constructor.
     * @param ScopeConfigInterface $config
     * @param ResolverInterface $resolver
     * @param Fulltext $fulltextHelper
     * @param string $fulltextSearchMode
     * @param array $preprocessors
     */
    public function __construct(ScopeConfigInterface $config,
                                ResolverInterface $resolver,
                                Fulltext $fulltextHelper,
                                string $fulltextSearchMode = Fulltext::FULLTEXT_MODE_BOOLEAN,
                                $preprocessors = [])
    {
        $this->config = $config;
        $this->fullTextHelper = $fulltextHelper;
        $this->resolverInterface = $resolver;
        $this->preprocessors = $preprocessors;
        $this->fullTextSearchMode = $fulltextSearchMode;
        parent::__construct($resolver, $fulltextHelper, $fulltextSearchMode, $preprocessors);
    }

    /**
     * @param ScoreBuilder $scoreBuilder
     * @param Select $select
     * @param RequestQueryInterface $query
     * @param string $conditionType
     * @return Select
     */
    public function build(
        ScoreBuilder $scoreBuilder,
        Select $select,
        RequestQueryInterface $query,
        $conditionType
    )
    {

        $this->specialCharacters = $this->config->getValue(self::SPECIAL_CHARACTERS_PATH);
        $this->minimalCharacterLength = (int)($this->config->getValue(self::MINIMAL_CHARACTER_LENGTH_PATH));
        $this->symbols = str_split($this->specialCharacters, 1);
        /** @var $query \Magento\Framework\Search\Request\Query\Match */
        $queryValue = $this->prepareQuery($query->getValue(), $conditionType);

        $fieldList = [];
        foreach ($query->getMatches() as $match) {
            $fieldList[] = $match['field'];
        }
        $resolvedFieldList = $this->resolverInterface->resolve($fieldList);

        $fieldIds = [];
        $columns = [];
        foreach ($resolvedFieldList as $field) {
            if ($field->getType() === FieldInterface::TYPE_FULLTEXT && $field->getAttributeId()) {
                $fieldIds[] = $field->getAttributeId();
            }
            $column = $field->getColumn();
            $columns[$column] = $column;
        }

        $matchQuery = $this->fullTextHelper->getMatchQuery(
            $columns,
            $queryValue,
            $this->fullTextSearchMode
        );
        $scoreBuilder->addCondition($matchQuery, true);

        if ($fieldIds) {
            $matchQuery = sprintf('(%s AND search_index.attribute_id IN (%s))', $matchQuery, implode(',', $fieldIds));
        }

        $select->where($matchQuery);

        return $select;
    }

    /**
     * Prepare query value for build function.
     *
     * @param string $queryValue
     * @param string $conditionType
     * @return string
     */
    protected function prepareQuery($queryValue, $conditionType)
    {
        $queryValue = str_replace($this->symbols, ' ', $queryValue);
        foreach ($this->preprocessors as $preprocessor) {
            $queryValue = $preprocessor->process($queryValue);
        }

        $stringPrefix = '';
        if ($conditionType === BoolExpression::QUERY_CONDITION_MUST) {
            $stringPrefix = '+';
        } elseif ($conditionType === BoolExpression::QUERY_CONDITION_NOT) {
            $stringPrefix = '-';
        }

        $queryValues = explode(' ', $queryValue);

        foreach ($queryValues as $queryKey => $queryValue) {
            if (empty($queryValue)) {
                unset($queryValues[$queryKey]);
            } else {
                $stringSuffix = $this->minimalCharacterLength > strlen($queryValue) ? '' : '*';
                $queryValues[$queryKey] = $stringPrefix . $queryValue . $stringSuffix;
            }
        }

        $queryValue = implode(' ', $queryValues);

        return $queryValue;
    }
}
